
/***************************************************************************
 *   This file is part of Aspect, a simple PEC tool.                       *
 *                                                                         *
 *   Copyright (C) 2007 by Wolfgang Hoffmann <woho@woho.de>                *
 *                                                                         *
 *   This program is free software, licensed under the GPL v2.             *
 *   See the file COPYING for more details.                                *
 ***************************************************************************/

#include "data.h"
#include "log.h"

#include <QtCore>
#include <math.h>




/********************************************************************/
/*!
    \class PecSample
    \brief Data container for one PEC table entry
    \sa PecData
*/

/*!
    \enum PecSample::Val
    Possible values for one PEC table entry
*/


int PecSample::val2Int(PecSample::Val eVal)
    {
    switch (eVal)
        {
        default:                return 0;
        case PecSample::faster: return 1;
        case PecSample::slower: return 2;
        case PecSample::reset:  return 3;
        }
    }


PecSample::Val PecSample::int2Val(int nVal)
    {
    switch (nVal)
        {
        default: return PecSample::keep;
        case 1:  return PecSample::faster;
        case 2:  return PecSample::slower;
        case 3:  return PecSample::reset;
        }
    }


int PecSample::val2Acc(PecSample::Val eVal)
    {
    switch (eVal)
        {
        default:                return  0;
        case PecSample::faster: return  1;
        case PecSample::slower: return -1;
        }
    }


PecSample::Val PecSample::acc2Val(int nAcc)
    {
    switch (nAcc)
        {
        default: return PecSample::keep;
        case  1: return PecSample::faster;
        case -1: return PecSample::slower;
        }
    }


PecSample::PecSample()
:   m_eVal(PecSample::keep),
    m_nAcc(0),
    m_nSpd(0),
    m_nPos(0)
    {
    }


PecSample::PecSample(PecSample::Val eVal)
:   m_eVal(eVal),
    m_nAcc(0),
    m_nSpd(0),
    m_nPos(0)
    {
    }


PecSample::PecSample(PecSample::Val eVal, int nAcc, int nSpd, int nPos)
:   m_eVal(eVal),
    m_nAcc(nAcc),
    m_nSpd(nSpd),
    m_nPos(nPos)
    {
    }




/********************************************************************/
/*!
    \class PecData
    \brief Data container for PEC table

    This class holds data for PEC (periodic error correction).
*/


PecData::PecData()
    {
    m_aTable.resize(c_nRowNum);
    clear();
    }


/*!
    Clear PEC table

    Sets first PEC value to PecSample::reset, all other to PecSample::keep.
*/

void PecData::clear()
    {
    m_nFactor = c_nFactorMin;
    for (int nRow = 0; nRow < c_nRowNum; nRow++)
        m_aTable[nRow] = PecSample();
    m_aTable[0].m_eVal = PecSample::reset;
    m_nOffset = 0;
    }


PecData::Traversal PecData::traversal() const
    {
    return m_eTraversal;
    }


void PecData::setTraversal(Traversal eDirection)
    {
    m_eTraversal = eDirection;
    recalcGeo();
    }


int PecData::factor() const
    {
    return m_nFactor;
    }


const PecSample &PecData::operator[](int nRow) const
    {
    static PecSample oNull;
    if ((nRow < 0) || (nRow >= c_nRowNum))
        return oNull;

    return m_aTable[nRow];
    }


PecSample::Val PecData::value(int nRow) const
    {
    return operator[](nRow).m_eVal;
    }


int PecData::valueInt(int nRow) const
    {
    return PecSample::val2Int(value(nRow));
    }


bool PecData::set(int nFactor, const QVector<PecSample::Val> &aTable)
    {
    if ((nFactor < c_nFactorMin) || (nFactor > c_nFactorMax))
        return false;
    if (aTable.size() != c_nRowNum)
        return false;

    m_nFactor = nFactor;
    for (int nRow = 0; nRow < c_nRowNum; nRow++)
        m_aTable[nRow].m_eVal = aTable[nRow];

    recalcGeo();
    return true;
    }


bool PecData::setFactor(int nFactor)
    {
    if ((nFactor < c_nFactorMin) || (nFactor > c_nFactorMax))
        return false;

    m_nFactor = nFactor;
    return true;
    }


bool PecData::setValue(int nRow, PecSample::Val eVal)
    {
    if ((nRow < 0) || (nRow >= c_nRowNum))
        return false;

    m_aTable[nRow].m_eVal = eVal;
    return true;
    }


void PecData::setFinish()
    {
    recalcGeo();
    }


/*!
    Calculate PEC table to acheive the given PE profile \a aProfile

    Background:

    PE is a positional error of the worm gear.

    Let \c x denote the rotation angle of the worm, running from \c 0 ... \c 256
    for one worm revolution (360°).<br>
    Let \c y denote the rotation angle of the worm wheel, running from \c 0 ... \c 256
    when advancing by one worm wheel tooth (360° / number of worm wheel teeth).

    Ideally, <tt>y = x + C</tt>, with C being an arbitrary constant
    (worm and worm wheel beeing endless, so there's no special start point).<br>
    With PE, <tt>y = x + e(x) + C</tt>.

    To compensate the PE, the MCU models e(x) with an approximation c(x), which has
    discrete slope values of <tt>-f, 0, f</tt>. The MCU holds a PEC table of 256 entries
    and a PEC factor, which describe <tt>c(x)</tt>. A PEC table entry has one of the
    following values:
    - \c 0: slope of c(x) is unchanged, i.e. <tt>c(x+1) - c(x) = c(x) - c(x-1)</tt>
    - \c 1: slope of c(x) increases, if slope is <= 0, otherwise is unchanged,
        i.e. <tt>c(x+1) - c(x) = c(x) - c(x-1) + f</tt>
    - \c 2: slope of c(x) decreases, if slope is >= 0, otherwise is unchanged,
        i.e. <tt>c(x+1) - c(x) = c(x) - c(x-1) - f</tt>
    - \c 3: slope of c(x) is 0, i.e. <tt>c(x) = c(x-1) = c(x+1)</tt>

    It seems that c(x) is dependent on the traversal direction, i.e. whether the
    PEC table is processed with increasing or decreasing x.

    Concerning \c f, empirical measurements with PEC factor 1 suggest a value of 0.0125

    \note Above description of PEC encoding is an empirical guess, which
        works for my settings. Some details or even the whole principle
        might be wrong for general usage ...

    \todo Currently the algorithm only considers a small range of values (the current,
        previous and next). That seems ok for smooth PE curves, but probably fails for
        PE with sharp jumps. So find an algorithm that looks at the whole picture and
        minimizes the mean square error.

    \sa recalcGeo()
*/

bool PecData::calc(const QVector<double> &aProfile)
    {
    LOG(4, "PecData::calc(): traversal %d, factor %d", (int)m_eTraversal, m_nFactor);
    if (aProfile.size() != c_nRowNum)
        {
        LOG(0, "PecData::calc(): aProfile has wrong size %d instead of %d",
            aProfile.size(), c_nRowNum);
        return false;
        }

    // traverse three times through the table:
    // - first time to get a correct start speed
    // - second time to fill in acceleration table
    // - third time as a consistency check if values are stable
    //   and to find minimum and maximum of resulting positions
    // this seems to work better than starting from a dedicated point
    // such as the profile minimum or maximum, which in theory has 0
    // profile slope, but practically can cause problems if we cannot
    // follow the profile correctly because the profile slope is too
    // steep for our maximum corrction speed
    int nSpd = 0;
    int nPos = (int)rint(aProfile[0]);
    int nPosMin = 1000;
    int nPosMax = -1000;
    for (int nRound = 0; nRound < 3; nRound++)
        {
        for (int nCnt = 0; nCnt < c_nRowNum; nCnt++)
            {
            // careful:
            // The firmware seems to model a PEC table entry not as a point,
            // but as an interval. The interval covers the area of the worm
            // where pecByte*4+pecBit is constant. Physical properties are
            // modeled like this:
            // - acceleration is a spike at the interval entry point, i.e.
            //   it's real position on the worm depends on traversal direction
            // - speed changes immediately at interval entry, since accel is a spike
            //   so speed is the same over the interval, independently of traversal
            //   direction, except if previous speed differs
            // - position reaches resulting value when exiting the interval,
            //   after moving from current to next point with current speed
            // So when looking at a PEC table entry point, it's values depend
            // on traversal direction:
            // - forward (this=n, prev=n-1, next=n+1):
            //   speed[this] = speed[prev] + acc[this]
            //   pos[this] = pos[prev] + speed[prev]
            // - backward (this=n, prev=n+1, next=n-1):
            //   speed[this] = speed[prev] + acc[next]
            //   pos[this] = pos[prev] + speed[prev]

            // nSpd and nPos are speed and position at entry of current interval
            int nRowThis = (m_eTraversal * nCnt + c_nRowNum) % c_nRowNum;
            int nRowNext = (m_eTraversal * (nCnt + 1) + c_nRowNum * 2) % c_nRowNum;
            int nRowAcc = (m_eTraversal == forward)? nRowThis: nRowNext;
            // dfProfile is position profile at half way between this and next point.
            // dfPosDiff is dfProfile minus reached position if continuing with current speed
            double dfMidProfile = (aProfile[nRowNext] + aProfile[nRowThis]) / 2.;
            double dfMidFactor = (double)m_nFactor * .5;
            double dfMidDiff = dfMidProfile - ((double)nPos + (double)nSpd * dfMidFactor);
            // calculate needed acceleration to minimize positional error
            int nAcc = 0;
            if (dfMidDiff > dfMidFactor)
                nAcc = 1;
            else if (dfMidDiff < -dfMidFactor)
                nAcc = -1;
            // adjust speed by acceleration and keep within limits
            nSpd += nAcc;
            if (nSpd < -1)
                nSpd = -1;
            if (nSpd > 1)
                nSpd = 1;
            // adjust position by speed
            nPos += nSpd * m_nFactor;
            LOG(4, "round %d, row %d: profile %.3f, pos %d, spd %d",
                nRound, nRowThis, dfMidProfile, nPos, nSpd);
            // depending on round, write acceleration into table
            // or perform consistency check
            if (nRound == 1)
                m_aTable[nRowAcc] = PecSample::acc2Val(nAcc);
            else if (nRound == 2)
                {
                if (nPosMin > nPos)
                    nPosMin = nPos;
                if (nPosMax < nPos)
                    nPosMax = nPos;
                if (nAcc != PecSample::val2Acc(m_aTable[nRowAcc].m_eVal))
                    LOG(0, "PecData::calc(): inconsistent result at row %d: %d != %d",
                        nRowAcc, nAcc, PecSample::val2Acc(m_aTable[nRowAcc].m_eVal));
                // set table to PecSample::reset if acc and spd are 0
                // strictly speaking, this is not necessary, but it shouldn't hurt, eighter.
                if ((nAcc == 0) && (nSpd == 0))
                    m_aTable[nRowAcc].m_eVal = PecSample::reset;
                }
            }
        }

    // Fill in speed and position profile from the generated acceleration sequence
    m_nOffset = (nPosMax + nPosMin) / 2;
    recalcGeo();

    return true;
    }


/*!
    Recalculate acceleration, speed and position from
    raw PEC table values
*/

void PecData::recalcGeo()
    {
    // traverse twice through the table:
    // - first time to get a correct start speed
    //   (necessary for tables without any PecSample::reset value)
    // - second time for storing acc, spd and pos
    int nSpd = 0;
    for (int nRound = 0; nRound < 2; nRound++)
        {
        int nPos = 0;
        for (int nCnt = 0; nCnt < c_nRowNum; nCnt++)
            {
            // careful:
            // The firmware seems to model a PEC table entry not as a point,
            // but as an interval. The interval covers the area of the worm
            // where pecByte*4+pecBit is constant. Physical properties are
            // modeled like this:
            // - acceleration is a spike at the interval entry point, i.e.
            //   it's real position on the worm depends on traversal direction
            // - speed changes immediately at interval entry, since accel is a spike
            //   so speed is the same over the interval, independently of traversal
            //   direction, except if previous speed differs
            // - position reaches resulting value when exiting the interval,
            //   after moving from current to next point with current speed
            // So when looking at a PEC table entry point, it's values depend
            // on traversal direction:
            // - forward (this=n, prev=n-1, next=n+1):
            //   speed[this] = speed[prev] + acc[this]
            //   pos[this] = pos[prev] + speed[prev]
            // - backward (this=n, prev=n+1, next=n-1):
            //   speed[this] = speed[prev] + acc[next]
            //   pos[this] = pos[prev] + speed[prev]

            // nSpd and nPos are speed and position at entry of current interval
            int nRowThis = (m_eTraversal * nCnt + c_nRowNum) % c_nRowNum;
            int nRowNext = (m_eTraversal * (nCnt + 1) + c_nRowNum * 2) % c_nRowNum;
            int nRowAcc = (m_eTraversal == forward)? nRowThis: nRowNext;
            int nAcc = 0;
            switch (m_aTable[nRowAcc].m_eVal)
                {
                case PecSample::reset:  nAcc =  0; nSpd = 0; break;
                case PecSample::faster: nAcc =  1; break;
                case PecSample::slower: nAcc = -1; break;
                case PecSample::keep:   nAcc =  0; break;
                }
            nSpd += nAcc;
            if (nSpd < -1)
                nSpd = -1;
            if (nSpd > 1)
                nSpd = 1;
            if (nRound == 0)
                continue;
            PecSample &rThis = m_aTable[nRowThis];
            rThis.m_nAcc = nAcc;
            rThis.m_nSpd = nSpd;
            rThis.m_nPos = nPos;
            nPos += nSpd * m_nFactor;
            }
        }

    // center positions such that m_nOffset is at half distance between minimum and maximum
    int nRowMin = 0;
    int nRowMax = 0;
    for (int nRow = 0; nRow < c_nRowNum; nRow++)
        {
        if (m_aTable.at(nRowMin).m_nPos > m_aTable.at(nRow).m_nPos)
            nRowMin = nRow;
        if (m_aTable.at(nRowMax).m_nPos < m_aTable.at(nRow).m_nPos)
            nRowMax = nRow;
        }
    int nPosMid = (m_aTable.at(nRowMin).m_nPos + m_aTable.at(nRowMax).m_nPos) / 2 - m_nOffset;
    for (int nRow = 0; nRow < c_nRowNum; nRow++)
        m_aTable[nRow].m_nPos -= nPosMid;
    }

